from urllib.parse import urljoin

import requests

from sharing_configs.models import SharingConfigsConfig

from .exceptions import ApiException


class SharingConfigsClient:
    def __init__(self) -> None:
        self.config = SharingConfigsConfig.get_solo()
        self.label = self.config.label
        label_url = f"{str(self.label)}/folder/"
        self.base_url = urljoin(self.config.api_endpoint, label_url)
        self.headers = {
            "content-type": "application/json",
            "authorization": f"Token {self.config.api_key}",
        }

    def get_list_folders_url(self) -> str:
        """url to get available folders and subfolders"""
        return self.base_url

    def get_folder_files_url(self, folder) -> str:
        """url to get files in a given folder"""
        return urljoin(self.base_url, f"{folder}/files/")

    def get_import_url(self, folder, filename) -> str:
        """url to get a given file"""
        return urljoin(self.base_url, f"{folder}/files/{filename}")

    def get_export_url(self, folder) -> str:
        """url to upload a file"""
        return urljoin(self.base_url, f"{folder}/files/")

    def export(self, folder, data) -> dict:
        """
        expect path required param folder
        """
        resp = requests.post(
            url=self.get_export_url(folder), headers=self.headers, json=data
        )
        try:
            resp.raise_for_status()
        except (requests.exceptions.HTTPError, requests.ConnectionError) as e:
            raise ApiException("Error during export of object")
        return resp.json()

    def import_data(self, folder: str, filename: str) -> bytes:
        """expect required path params: label,folder,filename to get binary data from API"""
        resp = requests.get(
            url=self.get_import_url(folder, filename), headers=self.headers
        )
        try:
            resp.raise_for_status()
        except (requests.exceptions.HTTPError, requests.ConnectionError) as e:
            raise ApiException("Error during import of object")
        return resp.content

    def get_folders(self, permission: dict) -> dict:
        """
        return dict with attr "results" containing list of folders
        """
        resp = requests.get(
            url=self.get_list_folders_url(), headers=self.headers, params=permission
        )
        try:
            resp.raise_for_status()
        except (requests.exceptions.HTTPError, requests.ConnectionError) as exc:
            raise ApiException({"error": "No folders"})

        return resp.json()

    def get_files(self, folder) -> dict:
        """
        expect required path param folder;
        return dict with attr "results" containing file names
        """
        resp = requests.get(self.get_folder_files_url(folder), headers=self.headers)
        try:
            resp.raise_for_status()
        except requests.HTTPError as exp:
            raise ApiException("No files available")
        return resp.json()



from django.db import models
from django.utils.translation import gettext_lazy as _

from solo.models import SingletonModel


class SharingConfigsConfig(SingletonModel):
    """
    Config for sharing
    """

    api_endpoint = models.URLField(
        _("api_endpoint"),
        max_length=250,
        help_text=_("Path to API point"),
    )
    api_key = models.CharField(
        _("api_key"),
        max_length=128,
        help_text=_("API key for authorization"),
    )
    label = models.CharField(
        _("label"),
        max_length=50,
        help_text=_("Label"),
    )
    default_organisation = models.CharField(
        _("default organisation"),
        max_length=100,
        blank=True,
        help_text=_("Default organisation"),
    )

    class Meta:
        verbose_name = _("Sharing configuration config")



# from logging import raiseExceptions
from django import forms
from django.utils.translation import gettext_lazy as _

from .exceptions import ApiException
from .utils import get_imported_folders_choices


class FolderForm(forms.Form):
    """
    Trigger an API call to get folders available for a given user
    """

    permission = None
    folder = forms.ChoiceField(label=_("Folders"), required=True, choices=[])

    def __init__(self, *args, **kwargs):
        """provide a list of folders(from API) for a drop-down menu based on permission.
        if api call fails raise custom exception"""

        folder_list = []
        try:
            folder_list = get_imported_folders_choices(self.permission)
            if len(folder_list) > 0:
                folder_list.insert(0, (None, "Choose a folder"))
                super().__init__(*args, **kwargs)
                self.fields["folder"].choices = list(folder_list)

        except ApiException as err:
            # If api call fails -> msg "help_text"
            folder_list.insert(0, (None, "Choose a folder"))
            super().__init__(*args, **kwargs)
            self.fields["folder"].help_text = "No folders available"
            self.fields["folder"].choices = folder_list


class ImportForm(FolderForm):
    """Provide form  with a list of readable folders"""

    permission = "read"
    file_name = forms.CharField(
        label=_("File name"),
        widget=forms.Select,
        required=True,
        help_text=_("Name of the file in storage folder"),
    )

    def clean_file_name(self):
        file_name = self.cleaned_data["file_name"]
        if file_name == "":
            raise forms.ValidationError("File should have a name")
        return file_name


class ExportToForm(FolderForm):
    """Provide form  with a list of writable folders"""

    permission = "write"

    file_name = forms.CharField(
        label=_("File name"),
        max_length=100,
        required=True,
        help_text=_("Name of the file in storage folder"),
    )
    overwrite = forms.BooleanField(
        label=_("Overwrite"),
        required=False,
        initial=False,
        help_text=_("Overwrite the existing file in the storage folder"),
    )



from typing import Union

from django.contrib import admin, messages
from django.contrib.auth import get_user_model
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import path, reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from solo.admin import SingletonModelAdmin

from sharing_configs.client_util import SharingConfigsClient
from sharing_configs.exceptions import ApiException
from sharing_configs.models import SharingConfigsConfig

from .exceptions import ApiException
from .forms import ExportToForm, ImportForm
from .utils import get_imported_files_choices


@admin.register(SharingConfigsConfig)
class SharingConfigsConfig(SingletonModelAdmin):
    pass


User = get_user_model()


class SharingConfigsExportMixin:
    """
    A class that prepares data and privides interface to make API call using credentials;
    The  get_sharing_configs_export_data() method raise NotImplementedError and should be
    overriden in a derived class.
    """

    change_form_template = "sharing_configs/admin/change_form.html"
    change_form_export_template = "sharing_configs/admin/export.html"
    sharing_configs_export_form = ExportToForm

    def get_sharing_configs_export_data(self, obj: object) -> Union[str, bytes]:
        """
        Derived class should provide object to export

        """
        raise NotImplemented

    def sharing_configs_export_view(self, request, object_id):
        """
        return template with form for GET request;
        process form data from POST request and make API call to endpoint
        """
        info = (
            self.model._meta.app_label,
            self.model._meta.model_name,
        )
        obj = self.get_object(request, object_id)
        initial = {"file_name": f"{obj.username}.json"}
        if request.method == "POST":
            form = self.get_sharing_configs_export_form(request.POST, initial=initial)
            if form.is_valid():
                author = request.user.username
                content = self.get_sharing_configs_export_data(obj)
                filename = form.cleaned_data.get("file_name")
                folder = form.cleaned_data.get("folder")
                data = {
                    "overwrite": form.cleaned_data.get("overwrite"),
                    "content": content,
                    "author": author,
                    "filename": filename,
                }

                obj_client = SharingConfigsClient()
                try:
                    resp = obj_client.export(folder, data)
                    msg = format_html(
                        _("The object {object} has been exported successfully"),
                        object=obj,
                    )
                    self.message_user(request, msg, level=messages.SUCCESS)
                    return redirect(
                        reverse(
                            f"admin:{info[0]}_{info[1]}_export",
                            kwargs={"object_id": obj.id},
                        )
                    )
                except ApiException as e:
                    msg = format_html(
                        _(f"Import of object failed: {e}"),
                    )
                    self.message_user(request, msg, level=messages.ERROR)

            if not form.is_valid():
                msg = format_html(
                    _("The object {object} has been not exported"),
                    object=obj,
                )
                self.message_user(request, msg, level=messages.ERROR)
            return render(
                request,
                self.import_template,
                {"form": form, "opts": self.model._meta},
            )

        else:
            form = self.sharing_configs_export_form(initial=initial)

        return render(
            request,
            self.change_form_export_template,
            {"object": obj, "form": form, "opts": obj._meta},
        )

    def get_urls(self):
        urls = super().get_urls()
        info = (
            self.model._meta.app_label,
            self.model._meta.model_name,
        )
        add_urls = [
            path(
                "<path:object_id>/export/",
                self.admin_site.admin_view(self.sharing_configs_export_view),
                name=f"{info[0]}_{info[1]}_export",
            ),
        ]

        return add_urls + urls

    def get_sharing_configs_export_form(self, *args, **kwargs):
        """return object export form"""
        if self.sharing_configs_export_form is not None:
            form = self.sharing_configs_export_form(*args, **kwargs)
            return form


class SharingConfigsImportMixin:
    """provide methods to download files(object) from storage using credentials"""

    change_list_template = "sharing_configs/admin/change_list.html"
    import_template = "sharing_configs/admin/import.html"
    sharing_configs_import_form = ImportForm

    def get_sharing_configs_import_data(
        self, filename: str, folder: str, author: str
    ) -> object:
        """
        Derived class should override params to import an object;
        Also need implementation to store a received object

        """
        raise NotImplemented

    def get_ajax_fetch_files(self, request, *args, **kwargs):
        """ajax call to pass chosen folder to a view"""
        folder = request.GET.get("folder_name")
        api_response_list_files = get_imported_files_choices(folder)
        if api_response_list_files:
            return JsonResponse({"resp": api_response_list_files, "status_code": 200})
        else:
            return JsonResponse({"status_code": 400, "error": "Unable to get folders"})

    def import_from_view(self, request, **kwargs):
        """
        return template with form and process data if form is bound;
        make API call to API point to download an object

        """
        info = (
            self.model._meta.app_label,
            self.model._meta.model_name,
        )
        if request.method == "POST":
            form = self.get_sharing_configs_import_form(request.POST)
            if form.is_valid():
                folder = form.cleaned_data.get("folder")
                filename = form.cleaned_data.get("file_name")
                obj = SharingConfigsClient()
                try:
                    resp_api_dict = obj.import_data(folder, filename)
                    msg = format_html(
                        _("The (file) object has been imported successfully!"),
                    )
                    self.message_user(request, msg, level=messages.SUCCESS)
                    return redirect(reverse(f"admin:{info[0]}_{info[1]}_import"))

                except ApiException as e:
                    msg = format_html(
                        _(f"Import of object failed: {e}"),
                    )
                    self.message_user(request, msg, level=messages.ERROR)

            if not form.is_valid():
                msg = format_html(
                    _("Something went wrong during object import"),
                )

                self.message_user(request, msg, level=messages.ERROR)

            return render(
                request,
                self.import_template,
                {"form": form, "opts": self.model._meta},
            )
        else:
            form = self.get_sharing_configs_import_form()
            return render(
                request,
                self.import_template,
                {
                    "form": form,
                    "opts": self.model._meta,
                },
            )

    def get_urls(self):
        urls = super().get_urls()
        info = (
            self.model._meta.app_label,
            self.model._meta.model_name,
        )

        add_urls = [
            path(
                "fetch/files/",
                self.admin_site.admin_view(self.get_ajax_fetch_files),
                name=f"{info[0]}_{info[1]}_ajax",
            ),
            path(
                "import/",
                self.admin_site.admin_view(self.import_from_view),
                name=f"{info[0]}_{info[1]}_import",
            ),
        ]

        return add_urls + urls

    def get_sharing_configs_import_form(self, *args, **kwargs):
        """return object of import form"""
        if self.sharing_configs_import_form is not None:
            form = self.sharing_configs_import_form(*args, **kwargs)
            return form



import requests

from sharing_configs.client_util import SharingConfigsClient

from .exceptions import ApiException


def get_folders_from_api(permission: str) -> dict:
    """
    make an API call to fetch data about folders with a given permission
    """
    obj = SharingConfigsClient()
    try:
        data = obj.get_folders(permission)
        return data
    except (requests.exceptions.HTTPError, requests.ConnectionError) as exc:
        raise ApiException({"error": "No folders"})


def get_imported_folders_choices(permission: str) -> list:
    """
    create list of tuples (folders name) based on api response
    ex:[('folder_one', 'folder_one'), ('folder_two', 'folder_two')]
    """
    folders_choices = []
    api_dict = get_folders_from_api(permission)
    results_list = api_dict.get("results", "No results in your dict")
    if results_list is not None:
        lst = FolderList()
        all_folders = lst.folder_collector(results_list)
        for folder in all_folders:
            folders_choices.append((folder, folder))
    else:
        folders_choices = []
    # folders_choices [('example_folder', 'example_folder'), ('example_subfolder', 'example_subfolder')]
    return folders_choices


def get_files_in_folder_from_api(folder: str) -> dict:
    """
    return files for a given folder
    """
    obj = SharingConfigsClient()
    try:
        content = obj.get_files(folder)
        return content
    except requests.exceptions.HTTPError as e:
        raise ApiException


def get_imported_files_choices(folder: str) -> list:
    """
    create list of filenames based on api response and to be passed to js

    """
    api_dict = get_files_in_folder_from_api(folder)
    results_list = api_dict.get("results", None)
    file_choices = []
    if results_list is not None:
        for item in results_list:
            file_choices.append(item.get("filename"))
    return file_choices


class FolderList:
    def __init__(self) -> None:
        self.folders_lst = []

    def folder_collector(self, lst) -> list:
        """
        Take a list and extract all (nested)folders from it.
        """

        for item in lst:
            self.folders_lst.append(item["name"])
            if len(item.get("children")) != 0:
                self.folder_collector(lst=item.get("children"))
        return self.folders_lst



# Generated by Django 4.0.3 on 2022-03-09 02:10

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SharingConfigsConfig",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "api_endpoint",
                    models.URLField(
                        help_text="Path to API point",
                        max_length=250,
                        verbose_name="api_endpoint",
                    ),
                ),
                (
                    "api_key",
                    models.CharField(
                        help_text="API key for authorization",
                        max_length=128,
                        verbose_name="api_key",
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        help_text="Label", max_length=50, verbose_name="label"
                    ),
                ),
                (
                    "default_organisation",
                    models.CharField(
                        blank=True,
                        help_text="Default organisation",
                        max_length=100,
                        verbose_name="default organisation",
                    ),
                ),
            ],
            options={
                "verbose_name": "Sharing configuration config",
            },
        ),
    ]
